/*
 * Mach Operating System
 * Copyright (c) 1992 Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/* Initially from the Mach Sun 3 distribution */

#include <mach/kern_return.h>
#include <mach/thread_status.h>
#include <mach/vm_param.h>

#include <kern/counters.h>
#include <kern/mach_param.h>
#include <kern/thread.h>
#include <kern/zalloc.h>

#include <machine/thread.h>
#include <machine/frame.h>

zone_t  pcb_zone;

void Thread_continue(void);
asm("_Thread_continue: movel d0,sp@(4); jra a2@");

void stack_attach(thread, stack, continuation)
        register thread_t       thread;
        vm_offset_t             stack;
        void                    (*continuation)();
{
  counter(if (++c_stacks_current > c_stacks_max)
                  c_stacks_max = c_stacks_current);

  thread->kernel_stack = stack;

  /*
   * We want to run continuation, giving it as an argument
   * the return value from Load_context/Switch_context.
   * Thread_continue takes care of the mismatch between
   * the argument-passing/return-value conventions.
   * This function will not return normally,
   * so we don't have to worry about a return address.
   *
   * We set up the stack like this:
   *
   *    sp -> Thread_continue: gets rts'd on by switch_context()
   *          dummy word: placeholder for return address for call to continuation
   *          uninitialized word, gets set to return parameter from switch_context() by Thread_continue()
   *          kernel context
   */
  {
    long *p = ((long*)STACK_KS(stack))-3;
    *p = (long)Thread_continue;
    STACK_KS(stack)->a2 = (long)continuation;
    STACK_KS(stack)->sp = (long)p;
  }
}

vm_offset_t stack_detach(register thread_t thread)
{
  register vm_offset_t    stack;

  counter(if (--c_stacks_current < c_stacks_min)
                  c_stacks_min = c_stacks_current);

  stack = thread->kernel_stack;
  thread->kernel_stack = 0;
  return stack;
}


/*
 *      stack_handoff:
 *
 *      Move the current thread's kernel stack to the new thread.
 */
void stack_handoff(thread_t old, thread_t new)
{
  fp_save(&old->pcb->fpu_state);

  if (old->task != new->task)
    PMAP_CONTEXT(vm_map_pmap(new->task->map), new);

  active_threads[0] = new;

  fp_restore(&new->pcb->fpu_state);
}

thread_t switch_context(old, continuation, new)
        register thread_t       old;
        void (*continuation)();
        register thread_t       new;
{
  struct kernel_state *old_state = STACK_KS(current_stack());

  fp_save(&old->pcb->fpu_state);

  if (old->task != new->task)
    PMAP_CONTEXT(vm_map_pmap(old->task->map), new);

  active_threads[0] = new;
  active_stacks[0] = new->kernel_stack;

  fp_restore(&new->pcb->fpu_state);

  {
    register void (*old_d0)() asm("d0") = old;

    if(continuation)
      {
        asm volatile(
          " movel %1,%0@;"
          " movel %0,%0@(12*4);"
          " moveml %2@,d2-d7/a2-a7;"
          " rts"
          :
          : "a" (((long*)old_state)-1),
            "r" (continuation),
            "a" (STACK_KS(new->kernel_stack)),
            "d" (old_d0));
      }
    else
      {
        asm("illegal");
        asm volatile(
          " pea 0f;"
          " moveml d2-d7/a2-a7,%1@;"
          " moveml %2@,d2-d7/a2-a7;"
          " rts;"
          "0:"
          : "=d" (old_d0)
          : "a" (old_state),
            "a" (STACK_KS(new->kernel_stack)),
            "d" (old_d0));
      }
    return(old_d0);
  }
}

volatile void
load_context(thread_t new)
{
  fp_restore(&new->pcb->fpu_state);

  active_threads[0] = new;
  active_stacks[0] = new->kernel_stack;

  asm volatile(
    " lea     _intstack+8192,sp;" /* Reset interrupt stack */
    " oriw    #0x3000,sr;" /* Force to master stack */
    " moveml  %0@,d2-d7/a2-a7;"
    " moveq   #0,d0;" /* No old thread */
    " rts"
    : : "a" (STACK_KS(new->kernel_stack)));
}

void
pcb_module_init()
{
  pcb_zone = zinit(sizeof(struct pcb),
                   THREAD_MAX * sizeof(struct pcb),
                   THREAD_CHUNK * sizeof(struct pcb),
                   FALSE, "mc680x0 pcb state");
}


extern thread_t fp_threads[];
extern short    fppstate;

void
pcb_init(thread_t thread)
{
  pcb_t          pcb;
  struct frame *regs;

  pcb = (pcb_t) zalloc(pcb_zone);
  thread->pcb = pcb;

  bzero((char *)pcb, sizeof(struct pcb));

  /*
   * Set up the initial user register block.
   * Put it at the very end of the user_state area,
   * so that a full-size exception frame will still
   * fit.
   */
  regs = ((struct frame*)&pcb->user_state[USER_STATE_SIZE]) - 1;
  pcb->user_regs = regs;

  /*
   *      Set up the initial user state.
   */
  regs->sr = SR_MASTER;   /* force to master stack on trap to kernel */
  regs->format = 0;       /* short format stack frame */
  regs->vector = 0;
}


void
pcb_terminate(thread_t thread)
{
  zfree(pcb_zone, (vm_offset_t) thread->pcb);
  thread->pcb = 0;
}


void
pcb_collect(thread_t thread)
{
}


/*
 *      thread_setstatus:
 *
 *      Set the status of the specified thread.
 */
kern_return_t thread_setstatus(thread, flavor, tstate, count)
        thread_t                thread;
        int                     flavor;
        thread_state_t          tstate;
        unsigned int            count;
{
  struct thread_state_regs *state;
  pcb_t pcb;
  struct frame *regs;

  if (flavor == THREAD_STATE_REGS)
    {
      if (count < THREAD_STATE_REGS_COUNT)
        return(KERN_INVALID_ARGUMENT);

      state = (struct sun_thread_state *) tstate;

      pcb = thread->pcb;
      regs = pcb->user_regs;

      bcopy(&state->d0,&regs->d0,sizeof(long)*16); /* FIXME: Will this work? */

      if (regs->pc != state->pc)
        {
          pcb->discard_frame = TRUE;
          regs->pc = state->pc;
        }

      /*
       *  Enforce user mode status register:
       *  must have user mode, user stack, interrupt priority 0.
       *  Force to master stack on trap to kernel mode.
       *  User may set trace bits, but both trace bits cannot be
       *  on (undefined).
       */
      regs->sr = state->sr & ~(SR_SUPERVISOR|SR_INT_MASK) | SR_MASTER;
      if (regs->sr & SR_TRACE_MASK == SR_T1|SR_T0)
          regs->sr &= ~SR_T0;

      if (state->fp.status == FPS_UNUSED)
        {
          bzero((char *)&pcb->fpu_state,
                sizeof(pcb->fpu_state));
        }
      else
        {
          /* FIXME: What do we do with the status? */
          pcb->fpu_state.status = state->fp;
        }

      return(KERN_SUCCESS);
    }
  else
    return(KERN_INVALID_ARGUMENT);
}


/*
 *      thread_getstatus:
 *
 *      Get the status of the specified thread.
 */
kern_return_t
thread_getstatus(thread, flavor, tstate, count)
        register thread_t       thread;
        int                     flavor;
        thread_state_t          tstate; /* pointer to OUT array */
        unsigned int            *count;         /* IN/OUT */
{
  register struct thread_state_regs *state;
  register pcb_t pcb;
  register struct frame *regs;

  switch (flavor)
    {
      case THREAD_STATE_FLAVOR_LIST:
        if (*count < 1)
          return(KERN_INVALID_ARGUMENT);
        tstate[0] = THREAD_STATE_REGS;
        *count = 1;
        break;

      case THREAD_STATE_REGS:
        if (*count < THREAD_STATE_REGS_COUNT) {
            return(KERN_INVALID_ARGUMENT);
        }

        state = (struct thread_state_regs*)tstate;

        pcb = thread->pcb;
        regs = pcb->user_regs;

        state->d0 = regs->d0; /* FIXME: Use bcopy??? */
        state->d1 = regs->d1;
        state->d2 = regs->d2;
        state->d3 = regs->d3;
        state->d4 = regs->d4;
        state->d5 = regs->d5;
        state->d6 = regs->d6;
        state->d7 = regs->d7;
        state->a0 = regs->a0;
        state->a1 = regs->a1;
        state->a2 = regs->a2;
        state->a3 = regs->a3;
        state->a4 = regs->a4;
        state->a5 = regs->a5;
        state->a6 = regs->a6;
        state->sp = regs->sp;

        state->pc = regs->pc;
        state->sr = regs->sr;

/*        if (fppstate > 0) { */
            state->fp = pcb->fpu_state.status;
            state->fp.status = pcb->fpu_state.istate.version == 0 ? FPS_UNUSED
              : pcb->fpu_state.istate.fsize <= sizeof(pcb->fpu_state.istate.u.idle882) ? FPS_IDLE
              : FPS_BUSY;
/*        }
        else
          bzero((char *)&state->fp, sizeof(state->fp));
*/

        *count = THREAD_STATE_REGS_COUNT;
        break;

      default:
        return(KERN_INVALID_ARGUMENT);
    }
  return(KERN_SUCCESS);
}

/*
 * Return prefered address of user stack.
 * Always returns low address.  If stack grows up,
 * the stack grows away from this address;
 * if stack grows down, the stack grows towards this
 * address.
 */
vm_offset_t
user_stack_low(vm_size_t stack_size)
{
  return(VM_MAX_ADDRESS - stack_size);
}

/*
 * Allocate argument area and set registers for first user thread.
 */
vm_offset_t
set_user_regs(vm_offset_t     stack_base,     /* low address */
              vm_offset_t     stack_size,
              int             *entry,
              vm_size_t       arg_size)
{
  vm_offset_t arg_addr;
  register struct frame *saved_state;

  arg_size = (arg_size + sizeof(int) - 1) & ~(sizeof(int)-1);
  arg_addr = stack_base + stack_size - arg_size;

  saved_state = current_thread()->pcb->user_regs;
  saved_state->sp = (int)arg_addr;
  saved_state->pc = entry[0];

  return (arg_addr);
}
